@@ -1,30 +1,30 @@
 // Tests run for (almost all) combinations of openssl, aessiv, plaintextnames.
 package matrix
 
-// File reading, writing, modification, truncate
+// File reading, writing, modification, truncate, ...
 //
-// Runs everything four times, for all combinations of
-// "-plaintextnames" and "-openssl".
-//
-// Test Matrix:
-//                       openssl=true openssl=false
-// plaintextnames=false  X            X
-// plaintextnames=true   X            X
+// Runs all tests N times, for the combinations of different flags specified
+// in the `matrix` variable.
 
 import (
 	"bytes"
 	"flag"
 	"fmt"
-	"io/ioutil"
+	"math/rand"
 	"os"
 	"os/exec"
+	"path/filepath"
 	"runtime"
+	"strings"
 	"sync"
 	"syscall"
 	"testing"
+	"time"
 
-	"github.com/rfjakob/gocryptfs/internal/syscallcompat"
-	"github.com/rfjakob/gocryptfs/tests/test_helpers"
+	"golang.org/x/sys/unix"
+
+	"github.com/rfjakob/gocryptfs/v2/internal/stupidgcm"
+	"github.com/rfjakob/gocryptfs/v2/tests/test_helpers"
 )
 
 // Several tests need to be aware if plaintextnames is active or not, so make this
@@ -36,48 +36,93 @@ type testcaseMatrix struct {
 	openssl        string
 	aessiv         bool
 	raw64          bool
+	extraArgs      []string
+}
+
+// isSet finds out if `extraArg` is set in `tc.extraArgs`
+func (tc *testcaseMatrix) isSet(extraArg string) bool {
+	for _, v := range tc.extraArgs {
+		if v == extraArg {
+			return true
+		}
+	}
+	return false
 }
 
 var matrix = []testcaseMatrix{
 	// Normal
-	{false, "auto", false, false},
-	{false, "true", false, false},
-	{false, "false", false, false},
+	{false, "auto", false, false, nil},
+	{false, "true", false, false, nil},
+	{false, "false", false, false, nil},
 	// Plaintextnames
-	{true, "true", false, false},
-	{true, "false", false, false},
+	{true, "true", false, false, nil},
+	{true, "false", false, false, nil},
 	// AES-SIV (does not use openssl, no need to test permutations)
-	{false, "auto", true, false},
-	{true, "auto", true, false},
+	{false, "auto", true, false, nil},
+	{true, "auto", true, false, nil},
 	// Raw64
-	{false, "auto", false, true},
+	{false, "auto", false, true, nil},
+	// -serialize_reads
+	{false, "auto", false, false, []string{"-serialize_reads"}},
+	{false, "auto", false, false, []string{"-sharedstorage"}},
+	{false, "auto", false, false, []string{"-deterministic-names"}},
+	// Test xchacha with and without openssl
+	{false, "true", false, true, []string{"-xchacha"}},
+	{false, "false", false, true, []string{"-xchacha"}},
 }
 
 // This is the entry point for the tests
 func TestMain(m *testing.M) {
 	// Make "testing.Verbose()" return the correct value
 	flag.Parse()
-	for _, testcase = range matrix {
+	var i int
+	for i, testcase = range matrix {
+		if testcase.openssl == "true" && stupidgcm.BuiltWithoutOpenssl {
+			continue
+		}
 		if testing.Verbose() {
 			fmt.Printf("matrix: testcase = %#v\n", testcase)
 		}
-		test_helpers.ResetTmpDir(!testcase.plaintextnames)
+		createDirIV := true
+		if testcase.plaintextnames {
+			createDirIV = false
+		} else if testcase.isSet("-deterministic-names") {
+			createDirIV = false
+		}
+		test_helpers.ResetTmpDir(createDirIV)
 		opts := []string{"-zerokey"}
+		//opts = append(opts, "-fusedebug")
 		opts = append(opts, fmt.Sprintf("-openssl=%v", testcase.openssl))
 		opts = append(opts, fmt.Sprintf("-plaintextnames=%v", testcase.plaintextnames))
 		opts = append(opts, fmt.Sprintf("-aessiv=%v", testcase.aessiv))
 		opts = append(opts, fmt.Sprintf("-raw64=%v", testcase.raw64))
+		opts = append(opts, testcase.extraArgs...)
 		test_helpers.MountOrExit(test_helpers.DefaultCipherDir, test_helpers.DefaultPlainDir, opts...)
+		before := test_helpers.ListFds(0, test_helpers.TmpDir)
+		t0 := time.Now()
 		r := m.Run()
+		if testing.Verbose() {
+			fmt.Printf("matrix: run took %v\n", time.Since(t0))
+		}
+		// Catch fd leaks in the tests. NOTE: this does NOT catch leaks in
+		// the gocryptfs FUSE process, but only in the tests that access it!
+		// All fds that point outside TmpDir are not interesting (the Go test
+		// infrastucture creates temporary log files we don't care about).
+		after := test_helpers.ListFds(0, test_helpers.TmpDir)
+		if len(before) != len(after) {
+			fmt.Printf("fd leak in test process? before, after:\n%v\n%v\n", before, after)
+			os.Exit(1)
+		}
 		test_helpers.UnmountPanic(test_helpers.DefaultPlainDir)
 		if r != 0 {
+			fmt.Printf("TestMain: matrix[%d] = %#v failed\n", i, testcase)
 			os.Exit(r)
 		}
 	}
 	os.Exit(0)
 }
 
-// Write "n" zero bytes to filename "fn", read again, compare hash
+// Write `n` random bytes to filename `fn`, read again, compare hash
 func testWriteN(t *testing.T, fn string, n int) string {
 	file, err := os.Create(test_helpers.DefaultPlainDir + "/" + fn)
 	if err != nil {
@@ -85,13 +130,17 @@ func testWriteN(t *testing.T, fn string, n int) string {
 	}
 
 	d := make([]byte, n)
-	written, err := file.Write(d)
-	if err != nil || written != len(d) {
-		t.Errorf("err=\"%s\", written=%d", err, written)
+	for i := range d {
+		// Fill with pattern
+		d[i] = byte(rand.Int())
+	}
+	_, err = file.Write(d)
+	if err != nil {
+		t.Fatal(err)
 	}
 	err = file.Close()
 	if err != nil {
-		t.Error(err)
+		t.Fatal(err)
 	}
 
 	test_helpers.VerifySize(t, test_helpers.DefaultPlainDir+"/"+fn, n)
@@ -124,9 +173,9 @@ func TestWrite100x100(t *testing.T) {
 	// Read and check 100 times to catch race conditions
 	var i int
 	for i = 0; i < 100; i++ {
-		hashActual := test_helpers.Md5fn(test_helpers.DefaultPlainDir + "/100")
+		hashActual := test_helpers.Md5fn(test_helpers.DefaultPlainDir + "/100x100")
 		if hashActual != hashWant {
-			fmt.Printf("Read corruption in loop #%d\n", i)
+			fmt.Printf("Read corruption in loop #%d: have=%s want=%s\n", i, hashActual, hashWant)
 			t.FailNow()
 		} else {
 			//fmt.Print(".")
@@ -134,6 +183,45 @@ func TestWrite100x100(t *testing.T) {
 	}
 }
 
+func TestWrite10Tight(t *testing.T) {
+	path := test_helpers.DefaultPlainDir + "/TestWrite10Tight"
+	content := make([]byte, 10)
+	buf := make([]byte, 100)
+	for i := 0; i < 100; i++ {
+		file, err := os.Create(path)
+		if err != nil {
+			t.Fatal(err)
+		}
+		_, err = file.Write(content)
+		if err != nil {
+			t.Fatal(err)
+		}
+		err = file.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
+		file, err = os.Open(path)
+		if err != nil {
+			t.Fatal(err)
+		}
+		n, err := file.Read(buf)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if n != 10 {
+			t.Fatalf("want 10 bytes, got %d", n)
+		}
+		err = file.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
+		err = os.Remove(path)
+		if err != nil {
+			t.Fatal()
+		}
+	}
+}
+
 // Hint for calculating reference md5sums:
 // dd if=/dev/zero count=1 bs=XYZ | md5sum
 func TestTruncate(t *testing.T) {
@@ -142,6 +230,7 @@ func TestTruncate(t *testing.T) {
 	if err != nil {
 		t.FailNow()
 	}
+	defer file.Close()
 	// Grow to two blocks
 	file.Truncate(7000)
 	test_helpers.VerifySize(t, fn, 7000)
@@ -193,157 +282,13 @@ func TestTruncate(t *testing.T) {
 	}
 }
 
-const (
-	// From man statfs
-	TMPFS_MAGIC      = 0x01021994
-	EXT4_SUPER_MAGIC = 0xef53
-)
-
-// isWellKnownFS decides if the backing filesystem is well-known.
-// The expected allocated sizes are only valid on tmpfs and ext4. btrfs
-// gives different results, but that's not an error.
-func isWellKnownFS(fn string) bool {
-	var fs syscall.Statfs_t
-	err := syscall.Statfs(fn, &fs)
-	if err != nil {
-		panic(err)
-	}
-	if fs.Type == EXT4_SUPER_MAGIC || fs.Type == TMPFS_MAGIC {
-		return true
-	}
-	return false
-}
-
-const FALLOC_DEFAULT = 0x00
-const FALLOC_FL_KEEP_SIZE = 0x01
-
-func TestFallocate(t *testing.T) {
-	if runtime.GOOS == "darwin" {
-		t.Skipf("OSX does not support fallocate")
-	}
-	fn := test_helpers.DefaultPlainDir + "/fallocate"
-	file, err := os.Create(fn)
-	if err != nil {
-		t.FailNow()
-	}
-	wellKnown := isWellKnownFS(test_helpers.DefaultCipherDir)
-	fd := int(file.Fd())
-	nBytes := test_helpers.Du(t, fd)
-	if nBytes != 0 {
-		t.Fatalf("Empty file has %d bytes", nBytes)
-	}
-	// Allocate 30 bytes, keep size
-	// gocryptfs ||        (0 blocks)
-	//      ext4 |  d   |  (1 block)
-	err = syscallcompat.Fallocate(fd, FALLOC_FL_KEEP_SIZE, 0, 30)
-	if err != nil {
-		t.Error(err)
-	}
-	var want int64
-	nBytes = test_helpers.Du(t, fd)
-	want = 4096
-	if nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	test_helpers.VerifySize(t, fn, 0)
-	// Three ciphertext blocks. The middle one should be a file hole.
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  h  |  d  |  (2 blocks)
-	// (Note that gocryptfs blocks are slightly bigger than the ext4 blocks,
-	// but the last one is partial)
-	err = file.Truncate(9000)
-	if err != nil {
-		t.Fatal(err)
-	}
-	nBytes = test_helpers.Du(t, fd)
-	want = 2 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	if md5 := test_helpers.Md5fn(fn); md5 != "5420afa22f6423a9f59e669540656bb4" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-	// Allocate the whole file space
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  d  |  d  |  (3 blocks
-	err = syscallcompat.Fallocate(fd, FALLOC_DEFAULT, 0, 9000)
-	if err != nil {
-		t.Fatal(err)
-	}
-	nBytes = test_helpers.Du(t, fd)
-	want = 3 * 4096
-	if nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	// Neither apparent size nor content should have changed
-	test_helpers.VerifySize(t, fn, 9000)
-	if md5 := test_helpers.Md5fn(fn); md5 != "5420afa22f6423a9f59e669540656bb4" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-
-	// Partial block on the end. The first ext4 block is dirtied by the header.
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  h  |  d  |  (2 blocks)
-	file.Truncate(0)
-	file.Truncate(9000)
-	nBytes = test_helpers.Du(t, fd)
-	want = 2 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	// Allocate 10 bytes in the second block
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  d  |  d  |  (3 blocks)
-	syscallcompat.Fallocate(fd, FALLOC_DEFAULT, 5000, 10)
-	nBytes = test_helpers.Du(t, fd)
-	want = 3 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	// Neither apparent size nor content should have changed
-	test_helpers.VerifySize(t, fn, 9000)
-	if md5 := test_helpers.Md5fn(fn); md5 != "5420afa22f6423a9f59e669540656bb4" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-	// Grow the file to 4 blocks
-	// gocryptfs |  h   |  h   |  d   |d|  (2 blocks)
-	//      ext4 |  d  |  d  |  d  |  d  | (4 blocks)
-	syscallcompat.Fallocate(fd, FALLOC_DEFAULT, 15000, 10)
-	nBytes = test_helpers.Du(t, fd)
-	want = 4 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	test_helpers.VerifySize(t, fn, 15010)
-	if md5 := test_helpers.Md5fn(fn); md5 != "c4c44c7a41ab7798a79d093eb44f99fc" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-	// Shrinking a file using fallocate should have no effect
-	for _, off := range []int64{0, 10, 2000, 5000} {
-		for _, sz := range []int64{0, 1, 42, 6000} {
-			syscallcompat.Fallocate(fd, FALLOC_DEFAULT, off, sz)
-			test_helpers.VerifySize(t, fn, 15010)
-			if md5 := test_helpers.Md5fn(fn); md5 != "c4c44c7a41ab7798a79d093eb44f99fc" {
-				t.Errorf("Wrong md5 %s", md5)
-			}
-		}
-	}
-	// Cleanup
-	file.Close()
-	syscall.Unlink(fn)
-	if !wellKnown {
-		// Even though most tests have been executed still, inform the user
-		// that some were disabled
-		t.Skipf("backing fs is not ext4 or tmpfs, skipped some disk-usage checks\n")
-	}
-}
-
 func TestAppend(t *testing.T) {
 	fn := test_helpers.DefaultPlainDir + "/append"
 	file, err := os.Create(fn)
 	if err != nil {
 		t.FailNow()
 	}
+	defer file.Close()
 	data := []byte("testdata123456789") // length 17
 	var buf bytes.Buffer
 	var hashWant string
@@ -377,25 +322,16 @@ func TestFileHoles(t *testing.T) {
 	if err != nil {
 		t.Errorf("file create failed")
 	}
+	defer file.Close()
 	foo := []byte("foo")
 	file.Write(foo)
 	file.WriteAt(foo, 4096)
-	_, err = ioutil.ReadFile(fn)
+	_, err = os.ReadFile(fn)
 	if err != nil {
 		t.Error(err)
 	}
 }
 
-// sContains - does the slice of strings "haystack" contain "needle"?
-func sContains(haystack []string, needle string) bool {
-	for _, element := range haystack {
-		if element == needle {
-			return true
-		}
-	}
-	return false
-}
-
 func TestRmwRace(t *testing.T) {
 
 	runtime.GOMAXPROCS(10)
@@ -405,10 +341,12 @@ func TestRmwRace(t *testing.T) {
 	if err != nil {
 		t.Fatalf("file create failed")
 	}
+	defer f1.Close()
 	f2, err := os.Create(fn)
 	if err != nil {
 		t.Fatalf("file create failed")
 	}
+	defer f2.Close()
 
 	oldBlock := bytes.Repeat([]byte("o"), 4096)
 
@@ -452,7 +390,7 @@ func TestRmwRace(t *testing.T) {
 		// but it must not be
 		// [oooooossss]
 
-		buf, _ := ioutil.ReadFile(fn)
+		buf, _ := os.ReadFile(fn)
 		m := test_helpers.Md5hex(buf)
 		goodMd5[m] = goodMd5[m] + 1
 
@@ -499,36 +437,64 @@ func TestFilenameEncryption(t *testing.T) {
 	}
 }
 
-// Test Mkdir and Rmdir
-func testMkdirRmdir(t *testing.T) {
-	test_helpers.TestMkdirRmdir(t, test_helpers.DefaultPlainDir)
-}
-
 // Test Rename
-func testRename(t *testing.T) {
+func TestRename(t *testing.T) {
 	test_helpers.TestRename(t, test_helpers.DefaultPlainDir)
 }
 
-// Overwrite an empty directory with another directory
-func TestDirOverwrite(t *testing.T) {
-	dir1 := test_helpers.DefaultPlainDir + "/DirOverwrite1"
-	dir2 := test_helpers.DefaultPlainDir + "/DirOverwrite2"
-	err := os.Mkdir(dir1, 0777)
+// Test that names of all lengths work
+func TestNameLengths(t *testing.T) {
+	f, err := os.Open(test_helpers.DefaultPlainDir)
 	if err != nil {
 		t.Fatal(err)
 	}
-	err = os.Mkdir(dir2, 0777)
+	entries, err := f.Readdirnames(0)
 	if err != nil {
 		t.Fatal(err)
 	}
-	err = syscall.Rename(dir1, dir2)
-	if err != nil {
-		t.Fatal(err)
+	f.Close()
+	cnt1 := len(entries)
+
+	wd := test_helpers.DefaultPlainDir + "/"
+	name := "x"
+	for len(name) < 2000 {
+		f, err := os.Create(wd + name + "x")
+		if err != nil {
+			break
+		}
+		name = name + "x"
+		f.Close()
+		f, err = os.Open(test_helpers.DefaultPlainDir)
+		if err != nil {
+			t.Fatal(err)
+		}
+		// In v1.7-rc2, we had a bug that allowed creation of too-long names.
+		// This threw errors in like this in READDIR:
+		//
+		//   OpenDir ".": invalid entry "gocryptfs.longname.wrE-izsR9ciEkP7JSCFDrk_d_Nj4mQo1dGY6hjuixAU=":
+		//   Could not read .name: ReadLongName: size=345 > limit=344
+		//
+		entries, err = f.Readdirnames(0)
+		if err != nil {
+			t.Fatal(err)
+		}
+		f.Close()
+		cnt2 := len(entries)
+		if cnt2 != cnt1+1 {
+			t.Fatalf("len=%d: expected %d dir entries, have %d: %v", len(name), cnt1+1, cnt2, entries)
+		}
+		err = syscall.Unlink(wd + name)
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+	if len(name) != 255 {
+		t.Errorf("maxlen=%d", len(name))
 	}
 }
 
 func TestLongNames(t *testing.T) {
-	fi, err := ioutil.ReadDir(test_helpers.DefaultCipherDir)
+	fi, err := os.ReadDir(test_helpers.DefaultCipherDir)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -541,40 +507,79 @@ func TestLongNames(t *testing.T) {
 		t.Fatalf("Could not create n255x: %v", err)
 	}
 	f.Close()
-	if !test_helpers.VerifyExistence(wd + n255x) {
+	if !test_helpers.VerifyExistence(t, wd+n255x) {
 		t.Errorf("n255x is not in directory listing")
 	}
-	// Rename long to long
+	// Rename long to long (target does not exist)
 	n255y := string(bytes.Repeat([]byte("y"), 255))
 	err = os.Rename(wd+n255x, wd+n255y)
 	if err != nil {
 		t.Fatalf("Could not rename n255x to n255y: %v", err)
 	}
-	if !test_helpers.VerifyExistence(wd + n255y) {
+	if !test_helpers.VerifyExistence(t, wd+n255y) {
 		t.Errorf("n255y is not in directory listing")
 	}
-	// Rename long to short
+	// Rename long to long (target exists)
+	f, err = os.Create(wd + n255x)
+	if err != nil {
+		t.Fatalf("Could not create n255x: %v", err)
+	}
+	f.Close()
+	err = os.Rename(wd+n255x, wd+n255y)
+	if err != nil {
+		t.Fatalf("Could not rename n255x to n255y: %v", err)
+	}
+	if !test_helpers.VerifyExistence(t, wd+n255y) {
+		t.Errorf("n255y is not in directory listing")
+	}
+	// Rename long to short (target does not exist)
 	err = os.Rename(wd+n255y, wd+"short")
 	if err != nil {
 		t.Fatalf("Could not rename n255y to short: %v", err)
 	}
-	if !test_helpers.VerifyExistence(wd + "short") {
+	if !test_helpers.VerifyExistence(t, wd+"short") {
 		t.Errorf("short is not in directory listing")
 	}
-	// Rename short to long
+	// Rename long to short (target exists)
+	f, err = os.Create(wd + n255y)
+	if err != nil {
+		t.Fatalf("Could not create n255y: %v", err)
+	}
+	f.Close()
+	err = os.Rename(wd+n255y, wd+"short")
+	if err != nil {
+		t.Fatalf("Could not rename n255y to short: %v", err)
+	}
+	if !test_helpers.VerifyExistence(t, wd+"short") {
+		t.Errorf("short is not in directory listing")
+	}
+	// Rename short to long (target does not exist)
 	err = os.Rename(wd+"short", wd+n255x)
 	if err != nil {
 		t.Fatalf("Could not rename short to n255x: %v", err)
 	}
-	if !test_helpers.VerifyExistence(wd + n255x) {
+	if !test_helpers.VerifyExistence(t, wd+n255x) {
 		t.Errorf("255x is not in directory listing II")
 	}
+	// Rename short to long (target exists)
+	f, err = os.Create(wd + "short")
+	if err != nil {
+		t.Fatalf("Could not create short: %v", err)
+	}
+	f.Close()
+	err = os.Rename(wd+"short", wd+n255x)
+	if err != nil {
+		t.Fatalf("Could not rename short to n255x: %v", err)
+	}
+	if !test_helpers.VerifyExistence(t, wd+n255x) {
+		t.Errorf("n255x is not in directory listing")
+	}
 	// Unlink
 	err = syscall.Unlink(wd + n255x)
 	if err != nil {
 		t.Fatalf("Could not unlink n255x: %v", err)
 	}
-	if test_helpers.VerifyExistence(wd + n255x) {
+	if test_helpers.VerifyExistence(t, wd+n255x) {
 		t.Errorf("n255x still there after unlink")
 	}
 	// Long symlink
@@ -583,7 +588,7 @@ func TestLongNames(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	if !test_helpers.VerifyExistence(wd + n255s) {
+	if !test_helpers.VerifyExistence(t, wd+n255s) {
 		t.Errorf("n255s is not in directory listing")
 	}
 	err = syscall.Unlink(wd + n255s)
@@ -601,7 +606,7 @@ func TestLongNames(t *testing.T) {
 		t.Error(err)
 	}
 	// Check for orphaned files
-	fi, err = ioutil.ReadDir(test_helpers.DefaultCipherDir)
+	fi, err = os.ReadDir(test_helpers.DefaultCipherDir)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -646,6 +651,9 @@ func TestLchown(t *testing.T) {
 
 // Set nanoseconds by path, symlink
 func TestUtimesNanoSymlink(t *testing.T) {
+	if runtime.GOOS == "darwin" {
+		t.Skipf("MacOS \"touch\" does not support \"--no-dereference\"")
+	}
 	path := test_helpers.DefaultPlainDir + "/utimesnano_symlink"
 	err := os.Symlink("/some/nonexisting/file", path)
 	if err != nil {
@@ -669,17 +677,15 @@ type utimesTestcaseStruct struct {
 	out [2]syscall.Timespec
 }
 
-func compareUtimes(want [2]syscall.Timespec, actual [2]syscall.Timespec) error {
-	tsNames := []string{"atime", "mtime"}
-	for i := range want {
-		if want[i].Sec != actual[i].Sec {
-			return fmt.Errorf("Wrong %s seconds: want=%d actual=%d", tsNames[i], want[i].Sec, actual[i].Sec)
-		}
-		if want[i].Nsec != actual[i].Nsec {
-			return fmt.Errorf("Wrong %s nanoseconds: want=%d actual=%d", tsNames[i], want[i].Nsec, actual[i].Nsec)
-		}
+// compareTimespec return true if the two passed Timespec are identical.
+func compareTimespec(want syscall.Timespec, actual syscall.Timespec) bool {
+	if want.Sec != actual.Sec {
+		return false
 	}
-	return nil
+	if want.Nsec != actual.Nsec {
+		return false
+	}
+	return true
 }
 
 const _UTIME_OMIT = ((1 << 30) - 2)
@@ -689,8 +695,8 @@ const _UTIME_OMIT = ((1 << 30) - 2)
 func doTestUtimesNano(t *testing.T, path string) {
 	utimeTestcases := []utimesTestcaseStruct{
 		{
-			in:  [2]syscall.Timespec{{Sec: 50, Nsec: 0}, {Sec: 50, Nsec: 0}},
-			out: [2]syscall.Timespec{{Sec: 50, Nsec: 0}, {Sec: 50, Nsec: 0}},
+			in:  [2]syscall.Timespec{{Sec: 50, Nsec: 0}, {Sec: 51, Nsec: 0}},
+			out: [2]syscall.Timespec{{Sec: 50, Nsec: 0}, {Sec: 51, Nsec: 0}},
 		},
 		{
 			in:  [2]syscall.Timespec{{Sec: 1, Nsec: 2}, {Sec: 3, Nsec: 4}},
@@ -705,19 +711,27 @@ func doTestUtimesNano(t *testing.T, path string) {
 			out: [2]syscall.Timespec{{Sec: 7, Nsec: 8}, {Sec: 5, Nsec: 6}},
 		},
 	}
+	if runtime.GOOS == "darwin" {
+		// darwin neither supports UTIME_OMIT nor nanoseconds (!?)
+		utimeTestcases = utimeTestcases[:1]
+	}
 	for i, tc := range utimeTestcases {
 		err := syscall.UtimesNano(path, tc.in[:])
 		if err != nil {
-			t.Fatal(err)
+			t.Fatalf("%q: %v", path, err)
 		}
 		var st syscall.Stat_t
 		err = syscall.Stat(path, &st)
 		if err != nil {
 			t.Fatal(err)
 		}
-		err = compareUtimes(tc.out, extractAtimeMtime(st))
-		if err != nil {
-			t.Errorf("Testcase %d: %v", i, err)
+		want := tc.out
+		have := extractAtimeMtime(st)
+		if !compareTimespec(want[0], have[0]) {
+			t.Errorf("Testcase %d: atime: want=%+v, have=%+v", i, want[0], have[0])
+		}
+		if !compareTimespec(want[1], have[1]) {
+			t.Errorf("Testcase %d: mtime: want=%+v, have=%+v", i, want[1], have[1])
 		}
 	}
 }
@@ -725,7 +739,7 @@ func doTestUtimesNano(t *testing.T, path string) {
 // Set nanoseconds by path, normal file
 func TestUtimesNano(t *testing.T) {
 	path := test_helpers.DefaultPlainDir + "/utimesnano"
-	err := ioutil.WriteFile(path, []byte("foobar"), 0600)
+	err := os.WriteFile(path, []byte("foobar"), 0600)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -734,11 +748,236 @@ func TestUtimesNano(t *testing.T) {
 
 // Set nanoseconds by fd
 func TestUtimesNanoFd(t *testing.T) {
+	if runtime.GOOS == "darwin" {
+		t.Skipf("MacOS does not have /proc")
+	}
 	path := test_helpers.DefaultPlainDir + "/utimesnanofd"
 	f, err := os.Create(path)
 	if err != nil {
 		t.Fatal(err)
 	}
+	defer f.Close()
 	procPath := fmt.Sprintf("/proc/self/fd/%d", f.Fd())
 	doTestUtimesNano(t, procPath)
 }
+
+// Make sure the Mknod call works by creating a fifo (named pipe)
+func TestMkfifo(t *testing.T) {
+	path := test_helpers.DefaultPlainDir + "/fifo1"
+	err := syscall.Mkfifo(path, 0700)
+	if err != nil {
+		t.Fatal(err)
+	}
+	path = test_helpers.DefaultPlainDir + "/gocryptfs.longname.XXX"
+	err = syscall.Mkfifo(path, 0700)
+	if err != nil {
+		t.Fatal(err)
+	}
+	err = os.Remove(path)
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+// TestMagicNames verifies that "magic" names are handled correctly
+// https://github.com/rfjakob/gocryptfs/issues/174
+func TestMagicNames(t *testing.T) {
+	names := []string{"warmup1", "warmup2", "gocryptfs.longname.QhUr5d9FHerwEs--muUs6_80cy6JRp89c1otLwp92Cs", "gocryptfs.diriv"}
+	for _, n := range names {
+		t.Logf("Testing n=%q", n)
+		p := test_helpers.DefaultPlainDir + "/" + n
+		// Create file
+		err := os.WriteFile(p, []byte("xxxxxxx"), 0200)
+		if err != nil {
+			t.Fatalf("creating file %q failed: %v", n, err)
+		}
+		// Rename magic to normal
+		err = os.Rename(p, test_helpers.DefaultPlainDir+"/x")
+		if err != nil {
+			t.Fatalf("rename 1 failed: %v", err)
+		}
+		// Rename normal to magic
+		err = os.Rename(test_helpers.DefaultPlainDir+"/x", p)
+		if err != nil {
+			t.Fatalf("rename 2 failed: %v", err)
+		}
+		// Unlink
+		err = syscall.Unlink(p)
+		if err != nil {
+			t.Fatal(err)
+		}
+		// Mkdir
+		err = os.Mkdir(p, 0700)
+		if err != nil {
+			t.Fatal(err)
+		}
+		// Rmdir
+		err = syscall.Rmdir(p)
+		if err != nil {
+			t.Fatal(err)
+		}
+		// Symlink
+		err = syscall.Symlink("xxxyyyyzzz", p)
+		if err != nil {
+			t.Fatal(err)
+		}
+		syscall.Unlink(p)
+		// Link
+		target := test_helpers.DefaultPlainDir + "/linktarget"
+		err = os.WriteFile(target, []byte("yyyyy"), 0600)
+		if err != nil {
+			t.Fatal(err)
+		}
+		err = syscall.Link(target, p)
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+}
+
+// Test that chmod works correctly
+func TestChmod(t *testing.T) {
+	path := test_helpers.DefaultPlainDir + "/" + t.Name()
+	file, err := os.Create(path)
+	if err != nil {
+		t.Fatal(err)
+	}
+	file.Close()
+	modes := []os.FileMode{0777, 0707, 0606, 0666, 0444, 0000, 0111, 0123, 0321}
+	for _, modeWant := range modes {
+		fi, err := os.Stat(path)
+		if err != nil {
+			t.Fatal(err)
+		}
+		err = syscall.Chmod(path, uint32(modeWant))
+		if err != nil {
+			t.Errorf("chmod %03o -> %03o failed: %v", fi.Mode(), modeWant, err)
+			continue
+		}
+		fi, err = os.Stat(path)
+		if err != nil {
+			t.Fatal(err)
+		}
+		modeHave := fi.Mode()
+		if modeHave != modeWant {
+			t.Errorf("modeHave %#o != modeWant %#o", modeHave, modeWant)
+		}
+	}
+}
+
+// Test that access(2) works correctly
+func TestAccess(t *testing.T) {
+	path := test_helpers.DefaultPlainDir + "/" + t.Name()
+	file, err := os.Create(path)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer file.Close()
+
+	err = unix.Access(path, unix.F_OK)
+	if err != nil {
+		t.Error(err)
+	}
+	err = unix.Access(path, unix.R_OK)
+	if err != nil {
+		t.Error(err)
+	}
+	err = unix.Access(path, unix.X_OK)
+	if err == nil {
+		t.Error("X_OK should have failed")
+	}
+}
+
+func TestStatfs(t *testing.T) {
+	var st syscall.Statfs_t
+	syscall.Statfs(test_helpers.DefaultPlainDir, &st)
+	if st.Bsize == 0 {
+		t.Errorf("statfs reports size zero: %#v", st)
+	}
+}
+
+// gocryptfs 2.0 reported the ciphertext size on symlink creation, causing
+// confusion: https://github.com/rfjakob/gocryptfs/issues/574
+func TestSymlinkSize(t *testing.T) {
+	p := filepath.Join(test_helpers.DefaultPlainDir, t.Name())
+	// SYMLINK reports the size to the kernel
+	if err := syscall.Symlink("foo", p); err != nil {
+		t.Fatal(err)
+	}
+	// Kernel serves us this value from the attr cache
+	var st syscall.Stat_t
+	if err := syscall.Lstat(p, &st); err != nil {
+		t.Fatal(err)
+	}
+	if st.Size != 3 {
+		t.Errorf("wrong size: have %d, want %d", st.Size, 3)
+	}
+}
+
+// gocryptfs 2.0+ reported the ciphertext size on hard link creation
+// https://github.com/rfjakob/gocryptfs/issues/724
+func TestLinkSize(t *testing.T) {
+	p := filepath.Join(test_helpers.DefaultPlainDir, t.Name()) + ".regular"
+	f, err := os.Create(p)
+	if err != nil {
+		t.Fatal(err)
+	}
+	_, err = f.WriteString("x")
+	f.Close()
+	if err != nil {
+		t.Fatal(err)
+	}
+	doTestLinkSize(t, p)
+
+	p = filepath.Join(test_helpers.DefaultPlainDir, t.Name()) + ".symlink"
+	err = syscall.Symlink("x", p)
+	if err != nil {
+		t.Fatal(err)
+	}
+	doTestLinkSize(t, p)
+}
+
+func doTestLinkSize(t *testing.T, p string) {
+	p2 := p + ".link"
+	err := syscall.Link(p, p2)
+	if err != nil {
+		t.Fatal(err)
+	}
+	var st syscall.Stat_t
+	err = syscall.Lstat(p2, &st)
+	if err != nil {
+		t.Fatal(filepath.Base(p2), err)
+	}
+	if st.Size != 1 {
+		t.Errorf("wrong %s size: want=1 have=%d", filepath.Base(p), st.Size)
+	}
+}
+
+// TestPwd check that /usr/bin/pwd works inside gocryptfs.
+//
+// This was broken in gocryptfs v2.0 with -sharedstorage:
+// https://github.com/rfjakob/gocryptfs/issues/584
+func TestPwd(t *testing.T) {
+	dir := test_helpers.DefaultPlainDir
+	for i := 0; i < 3; i++ {
+		cmd := exec.Command("pwd")
+		cmd.Dir = dir
+		out, err := cmd.CombinedOutput()
+		if err != nil {
+			t.Log(strings.TrimSpace(string(out)))
+			t.Fatalf("dir %q: %v", dir, err)
+		}
+		dir = dir + "/" + t.Name()
+		os.Mkdir(dir, 0700)
+	}
+}
+
+// TestRootIno checks that inode number of the root dir is set
+// https://github.com/hanwen/go-fuse/issues/399
+func TestRootIno(t *testing.T) {
+	var st syscall.Stat_t
+	syscall.Stat(test_helpers.DefaultPlainDir, &st)
+	if st.Ino == 0 {
+		t.Errorf("inode number of root dir is zero")
+	}
+}
